#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
AETVGen.py: Generate test vectors for the Authenticated Encryption algorithms according
            to protocol specified in.

Syntax Note:  "_WIDTH" always have the unit of bit(s)
              "_SIZE"  always have the unit of byte(s)
              "_WORD"  refers to the number of WORDS for a given widthData
"""

__author__     = "Ekawat (Ice) Homsirikamol"
__copyright__  = "Copyright (c) 2015 CERG @ GMU, USA"
__license__    = "GPL 3.0"

# Core imports
import random
import sys
import os
import binascii
import math

# ===================================
# CFFI - External library declaration
# ===================================
import cffi
ffi = cffi.FFI()
ffi.cdef("""
    int crypto_aead_encrypt(
        unsigned char *c,unsigned long long *clen,
        const unsigned char *m,unsigned long long mlen,
        const unsigned char *ad,unsigned long long adlen,
        const unsigned char *nsec,
        const unsigned char *npub,
        const unsigned char *k
	);
    int crypto_aead_decrypt(
        unsigned char *m,unsigned long long *mlen,
        unsigned char *nsec,
        const unsigned char *c,unsigned long long clen,
        const unsigned char *ad,unsigned long long adlen,
        const unsigned char *npub,
        const unsigned char *k
	);
""")

# ====================
# Algorithms and paths
# ====================

LIB_DIR = os.path.realpath("../lib")
if os.name == "nt":
    # windows
    try:
        os.environ['path'] += os.pathsep + os.path.realpath("../lib")
    except KeyError:
        os.environ['path'] = os.path.realpath("../lib")

else:
    # linux
    try:
        os.environ['LD_LIBRARY_PATH'] += os.pathsep + os.path.realpath("../lib")
    except KeyError:
        os.environ['LD_LIBRARY_PATH'] = os.path.realpath("../lib")

class AETVGenError(Exception):
    """Class of error used in the AE implementation
    Use: raise AETVGenError.AETVGenError("Text to be displayed")"""
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class AETVgen:
    """
    Class for generating test vector for
    Authenticated Encryption with Authenticated Data (AEAD)
    """

    # Constant declarations
    INS_PREFIX = 'INS'
    HDR_PREFIX = 'HDR'
    DAT_PREFIX = 'DAT'

    OPCODE_ENCRYPT      = 0
    OPCODE_DECRYPT      = 1
    OPCODE_AEAD_ENCRYPT = 2
    OPCODE_AEAD_DECRYPT = 3
    OPCODE_LOAD_KEY     = 4
    OPCODE_LOAD_RKEY    = 5
    OPCODE_ACT_KEY      = 7
    OPCODE_MAC          = 8
    OPCODE_HASH         = 9
    OPCODE_PRNG         = 10
    OPCODE_AE_PASS      = 14
    OPCODE_AE_FAIL      = 15
    CODE_OPCODE = {OPCODE_ENCRYPT:'Encrypt', OPCODE_DECRYPT:'Decrypt', OPCODE_AEAD_ENCRYPT:'AEAD encrypt', OPCODE_AEAD_DECRYPT:'AEAD decrypt',
                   OPCODE_LOAD_KEY:'Load key', OPCODE_LOAD_RKEY:'Load round key', OPCODE_ACT_KEY:'Activate key', OPCODE_MAC:'MAC', OPCODE_HASH:'HASH', OPCODE_PRNG:'PRNG',
                   OPCODE_AE_PASS:'Authentication passed', OPCODE_AE_FAIL:'Authentication failed'}

    STYPE_RESERVED      = 0
    STYPE_NPUB          = 1
    STYPE_AD            = 2
    STYPE_PLAINTEXT     = 3
    STYPE_CIPHERTEXT    = 4
    STYPE_TAG           = 5
    STYPE_KEY           = 6
    STYPE_ROUNDKEY      = 7
    STYPE_NSEC          = 8
    STYPE_ENC_NSEC      = 9
    STYPE_LEN           = 12
    CODE_STYPE = {STYPE_RESERVED:'Reserved',     STYPE_NPUB:'Npub',
                  STYPE_AD:'Associated Data',    STYPE_PLAINTEXT:'Plaintext',
                  STYPE_CIPHERTEXT:'Ciphertext', STYPE_TAG:'Tag',
                  STYPE_KEY:'Key',               STYPE_ROUNDKEY:'Round key',
                  STYPE_NSEC:'Nsec' ,            STYPE_ENC_NSEC:'Encrypted Nsec',
                  STYPE_LEN:'Length'}

    FILE_ALL            = 0
    FILE_INPUT          = 1
    FILE_OUTPUT         = 2
    FILE_PUBLIC         = 3
    FILE_PUBLIC_INPUT   = 4
    FILE_SECRET         = 5
    FILE_C              = 6
    FILE_VHDL           = 7

    def __init__(self, **kwargs):
        self.param = dict()
        # Default value of optional settings
        self.param['caesarLib']      = 'aes128gcmv1'
        self.param['testMode']       = 0        # Mode for fixing the message and key value to one block
        self.param['sizeKey']        = 16
        self.param['enableRoundKey'] = False
        self.param['sizeRoundKey']   = 16
        self.param['totalRoundKey']  = 11
        self.param['sizeNpub']       = 12
        self.param['enableNsec']     = False
        self.param['sizeNsec']       = 16
        self.param['sizeTag']        = 16
        self.param['blockSize']      = 16
        self.param['blockSizeAD']    = 16
        self.param['enableLoadLen']  = False
        self.param['ctrSizeAd']      = 8
        self.param['ctrSizeD']       = 8
        self.param['dontCare']       = False
        # Test vector format
        self.param['plainTextMode']  = 0        # Default write mode (1 = No Npub segment with AD = AD + Npub)
        self.param['cipherTextMode'] = 0        # Default write mode (1 = No Npub segment with AD = AD + Npub)
        self.param['reverseDblk']    = False    # Reverse block order for decryption
        self.param['maxSizeSegment'] = 100000
        self.param['padAD']          = 0        # pad Authenticated Data block mode.
        self.param['padD']           = 0        # pad Data block mode.                        
        # Input Message Parameters
        self.param['minSizeAD']      = 0
        self.param['maxSizeAD']      = 512
        self.param['minSize']        = 0
        self.param['maxSize']        = 512
        # Hardware I/O width
        self.param['sizePIO']        = 4
        self.param['sizeSIO']        = 4
        # Debugging options
        self.param['startTV']        = 0
        self.param['verbose']        = False
        self.param['decrypt']        = False    # Perform decryption after encryption (Only activated when 'verbose' is True)
        # Output files
        self.param['filePDI']        = 'pdi.txt'
        self.param['fileSDI']        = 'sdi.txt'
        self.param['fileDO']         = 'do.txt'
        self.param['fileC']          = 'hls_tb.c'
        self.param['fileVhdl']       = 'CipherCoreTest.vhd'
        
        # [Experimental faeatures]
        # CipherCore parameters
        self.param['genCipherCoreTvC']      = False
        self.param['genCipherCoreTvVhdl']   = False
        #       Implementation parameters (applicable when genCipherCoreTv* is active)
        self.param['padStyle']       = 1        # padding style
        self.param['Latency']        = {'EncTagCycles': 0, 'DecTagCycles': 0, 'NoMsgTagCycles': 0}
        #       C related paramter
        self.param['sizeElement']    = 8
        # Undocumented parameters (future support)
        self.param['algorithm']      = 'caesar'
        self.param['stdout']         = False
        self.param['libVer']         = None     # Library version (Unused)
        self.param['opcode']         = False
        self.param['minSizeRNG']     = 1
        self.param['maxSizeRNG']     = 512
        self.ABYTES                  = 100      # Default additional bytes (Used for memory allocation)
        """
        PLAINTEXT_MODE:
        {0} N_A_M: Separate Nonce, Associated Data, and Message segments.
        {1} NA_M: The Associated Data segment contains Nonce concatenated with Associated Data.
        {2} AN_M: The Associated Data segment contains Associated Data concatenated with Nonce.
        {3} N_A_M_A: Separate Nonce, Associated Data - Header, Message, and Associated Data - Trailer segments.

        CIPHERTEXT_MODE:
        {0} C_T: Separate Ciphertext and Tag segments.
        {1} CT : The Ciphertext segment contains Ciphertext concatenated with Tag.
        {2} Cexp_T: Separate expanded Ciphertext

        When
            CIPHERTEXT_MODE = CT
        then
        1) during encryption
            a. len_d = |M|
            b. The tag output of the Datapath is not used
            c. The output processor does not wait for 1 at the tag_write output of the Datapath
            d. The size of C in the ciphertext segment header = |M|+|T|

        2) during decryption
            a. The size of C in the ciphertext segment header = |M|+|T|
            b. len_d = |M|+|T| (|M| is calculated inside of the datapath)
            c. The exp_tag input of the Datapath is not used.
            d. The exp_tag_ready input of the Controller is not used.

        When
            CIPHERTEXT_MODE = C_T
        then
        1) during encryption
            a. len_d = |M|
            b. The tag output of the Datapath is used
            c. The output processor waits for 1 at the tag_write output of the Datapath
            d. The size of C in the ciphertext segment header = |M|

        2) during decryption
            a. The size of C in the ciphertext segment header = |M|
            b. len_d = |M|
            c. The exp_tag input of the Datapath is used.
            d. The exp_tag_ready input of the Controller is used.

        When
            CIPHERTEXT_MODE = Cexp_T
        then
        1) during encryption
            a. len_d = |M|
            b. The tag output of the Datapath is used
            c. The output processor waits for 1 at the tag_write output of the Datapath
            d. The size of C in the ciphertext segment header = |M| but the output processor expects n*ceil(|M|/n) bits of the ciphertext

        2) during decryption
            a. The size of C in the ciphertext segment header = |M|,
               but the input processor reads and passes to the Datapath n*ceil(|M|/n) bits of the ciphertext
            b. len_d = |M|
            c. The exp_tag input of the Datapath is used.
            d. The exp_tag_ready input of the Controller is used.
        """

        for key, value in kwargs.items():
            if key.lower()              == 'stdout'.lower():
                self.param['stdout']     = value
            elif key.lower()              == 'testMode'.lower():
                if value not in [0, 1, 2, None]:
                    raise AETVGenError.AETVGenError("Invalid 'testMode' option: {}".format(value))
                self.param['testMode'] = value
                if value > 0:
                    print(" -- Entering test mode -- ");
            elif key.lower()            == 'blockSize'.lower():
                self.param['blockSize'] = int(value)
            elif key.lower()            == 'blockSizeAD'.lower():
                self.param['blockSizeAD'] = int(value)
            elif key.lower()            == 'caesarLib'.lower():
                self.param['caesarLib'] = value
            elif key.lower()            == 'cipherTextMode'.lower():
                self.param['cipherTextMode'] = value
            elif key.lower()            == 'decrypt'.lower():
                self.param['decrypt'] = value
            elif key.lower()            == 'dontCare'.lower():
                self.param['dontCare'] = value
            elif key.lower()            == 'enableLoadLen'.lower():
                self.param['enableLoadLen'] = value
            elif key.lower()            == 'ctrSizeAd'.lower():
                self.param['ctrSizeAd'] = value
            elif key.lower()            == 'ctrSizeD'.lower():
                self.param['ctrSizeD']  = value
            elif key.lower()            == 'Latency'.lower():
                self.param['Latency']   = value
            elif key.lower()            == 'genCipherCoreTvC'.lower():
                self.param['genCipherCoreTvC'] = value
            elif key.lower()            == 'sizeElement'.lower():
                self.param['sizeElement'] = value
            elif key.lower()            == 'genCipherCoreTvVhdl'.lower():
                self.param['genCipherCoreTvVhdl'] = value
            elif key.lower()            == 'sizeKey'.lower():
                self.param['sizeKey']   = value
            elif key.lower()            == 'sizeRoundKey'.lower():
                self.param['sizeRoundKey']   = value
            elif key.lower()            == 'sizeNpub'.lower():
                self.param['sizeNpub']  = value
            elif key.lower()            == 'sizeTag'.lower():
                self.param['sizeTag']   = value
            elif key.lower()            == 'minSize'.lower():
                self.param['minSize']  = value
            elif key.lower()            == 'maxSize'.lower():
                self.param['maxSize']  = value
            elif key.lower()            == 'minSizeAD'.lower():
                self.param['minSizeAD'] = value
            elif key.lower()            == 'maxSizeAD'.lower():
                self.param['maxSizeAD'] = value
            elif key.lower()            == 'maxSizeSegment'.lower():
                self.param['maxSizeSegment'] = value
            elif key.lower()            == 'opcode'.lower():
                self.param['opcode'] = value
            elif key.lower()            == 'padAD'.lower():
                self.param['padAD'] = value
            elif key.lower()            == 'padD'.lower():
                self.param['padD'] = value
            elif key.lower()            == 'padStyle'.lower():
                self.param['padStyle'] = value
            elif key.lower()            == 'plainTextMode'.lower():
                self.param['plainTextMode'] = value
            elif key.lower()            == 'reverseDblk'.lower():
                self.param['reverseDblk'] = value
            elif key.lower()            == 'sizeNsec'.lower():
                self.param['sizeNsec'] = value
            elif key.lower()            == 'sizePIO'.lower():
                self.param['sizePIO'] = value
            elif key.lower()            == 'sizeSIO'.lower():
                self.param['sizeSIO'] = value
            elif key.lower()            == 'verbose'.lower():
                self.param['verbose'] = value
            elif key.lower()            == 'totalRoundKey'.lower():
                self.param['totalRoundKey'] = value
            elif key.lower()            == 'startTV'.lower():
                self.param['startTV'] = value
            elif key.lower()            == 'enableNsec'.lower():
                self.param['enableNsec'] = True if (value == 1 or value == True) else False
            elif key.lower()            == 'enableRoundKey'.lower():
                self.param['enableRoundKey'] = True if (value == 1 or value == True) else False
            else:
                print("Key -> {}".format(key))
                raise AETVGenError.AETVGenError("Invalid init argument, unknown key -> {}".format(key))

        if (self.param['maxSizeSegment'] % self.param['sizePIO'] > 0) or \
           (self.param['maxSizeSegment'] % self.param['sizeSIO'] > 0):
            raise ValueError("Maximum segment size has to be a multiple of I/O width")

        # Default settings
        self.pdi                              = ''
        self.pdo                              = ''
        self.do                               = ''
        self.msgid                            = 0
        self.keyid                            = 0
        self.key                              = ''
        self.blockSize                        = self.param['blockSize']
        self.blockSizeAD                      = self.param['blockSizeAD']

        #### Adjust CAESAR's library name to target a specific file based on system and the verbose option.
        if self.param['algorithm'].lower() in ['caesar']:
            if self.param['verbose'] == True:
                self.param['caesarLib']    = "{}_verbose".format(self.param['caesarLib'])
            if os.name == "nt":
                self.param['caesarLib']    = "{}.dll".format(self.param['caesarLib'])
            else:
                self.param['caesarLib']    = "{}/{}.so".format(os.path.realpath("../lib"), self.param['caesarLib'])
            self.caesar = ffi.dlopen(self.param['caesarLib'])
        else:
            raise ValueError("CAESAR must be used as the first input parameter.")

    def setParam(self, param, value):
        self.param[param] = value

    def getHdrIns(self, opcode, iowidth):
        """
        Format of instruction header is as followed:
        8-bit   message ID
        4-bit   zeros
        4-bit   opcode
        8-bit   key ID
        others  zeros
        """
        # Set lower bound for header width
        if iowidth < 4:
            iowidth = 4
        ins    = "{0:02X}{z:01X}{1:01X}{2:02X}{z:{z}{fill}X}".format(self.msgid,opcode,self.keyid, z=0, fill=self._calcIOFill(6, iowidth))
        return ins

    def getHdrCmd(self, segType, pad, eoi, eot, size, iowidth):
        """
        Format of command header is as followed:
        8-bit   message ID
        4-bit   segment type
        1-bit   reserved
        1-bit   padding flag
        1-bit   last segment of message flag
        1-bit   last segment of segment type flag
        others  size
        """
        # Set lower bound for header width
        if iowidth < 4:
            iowidth = 4
        if segType in [self.STYPE_KEY, self.STYPE_ROUNDKEY]:
            id = self.keyid
        else:
            id = self.msgid
        nibble = int("0{0}{1}{2}".format(pad,eoi,eot), 2)
        hdr    = "{0:02X}{1:01X}{2:01X}{3:{z}{fill}X}".format(id,segType,nibble,int(size), z=0, fill=self._calcIOFill(4, iowidth))
        return hdr

    def compute(self, algorithm, opcode, key, npub, ad, sizeAD, msg, sizeMsg, nsec):
        (encrypted, sizeEncrypted, tag) = ('',sizeMsg,'')

        if algorithm.lower() in ['aes_cmac', 'aes-cmac']:
            # key = int('0x' + key,16)
            # msg = int('0x' + msg,16)
            # key = long_to_bytes(key)
            # msg = long_to_bytes(msg)
            alg = aes.AES_CMAC(key)
            tag = alg.digest(msg)
            tag = "{0:0{size}X}".format(bytes_to_long(tag), size=self.param['sizeTag']*2)
        elif algorithm.lower() in ['aes_ctr', 'aes-ctr']:
            alg = aes.AES_CTR(key, npub)
            encrypted = alg.encrypt(msg.decode("hex"))
            encrypted = encrypted.encode("hex").upper()
        # elif algorithm.lower() in ['aes_ecb', 'aes-ecb']:
            # key  = int('0x' + key,16)
            # msg  = int('0x' + msg,16)
            # key = long_to_bytes(key)
            # msg = long_to_bytes(msg)
            # alg = AES.AES_ECB(key)
            # encrypted = alg.encrypt(msg)
            # encrypted = "{0:0{size}X}".format(bytes_to_long(encrypted), size=sizeMsg*2)
        elif algorithm.lower() in ['aes_fortuna', 'aes-fortuna']:
            alg = AES.AES_FORTUNA(key, npub)
            encrypted = alg.generate(sizeMsg).encode("hex").upper()
        elif algorithm.lower() in ['aes_hash', 'aes-hash']:
            alg = AES.AES_HASH()
            tag = alg.digest(msg)
        elif    algorithm.lower() in ['aes_ocb3', 'aes-ocb3']:
            key  = binascii.unhexlify(key)
            npub = binascii.unhexlify(npub)
            ad   = binascii.unhexlify(ad)
            msg  = binascii.unhexlify(msg)
            # Allocate space for output
            c    = "0"*sizeMsg + "0"*self.param['sizeTag']
            OCB_ENCRYPT = 1
            OCB_DECRYPT = 0
            self.aes_ocb3.ocb_crypt(c, key, npub, ad, sizeAD, msg, sizeMsg, OCB_ENCRYPT)

            encrypted = "".join(c[i].encode('hex').upper() for i in range(0,sizeMsg))
            tag       = "".join(c[i].encode('hex').upper() for i in range(sizeMsg,sizeMsg+self.param['sizeTag']))
        elif algorithm.lower() in ['caesar']:   # CAESAR Interface
            ### Input

            ## Add additional 0000s to input data buffer to prevent any weird problems
            msg  = msg  + "0"*(32-len(msg))  if (len(msg)%32 > 0)  else msg
            ad   = ad   + "0"*(32-len(ad))   if (len(ad)%32 > 0)   else ad
            nsec = nsec + "0"*(32-len(nsec)) if (len(nsec)%32 > 0) else nsec
            npub = npub + "0"*(32-len(npub)) if (len(npub)%32 > 0) else npub
            key  = key  + "0"*(32-len(key))  if (len(key)%32 > 0)  else key

            m       = ffi.new("char[]", binascii.unhexlify(msg))
            mlen    = ffi.cast("unsigned long long", sizeMsg)
            c       = ffi.new("char[]", sizeMsg+self.ABYTES)    # Add more buffers by default
            clen    = ffi.new("unsigned long long *", sizeMsg+self.ABYTES)
            ad      = ffi.new("char[]", binascii.unhexlify(ad))
            adlen   = ffi.cast("unsigned long long", sizeAD)
            nsec    = ffi.new("char[]", binascii.unhexlify(nsec))
            npub    = ffi.new("char[]", binascii.unhexlify(npub))
            key     = ffi.new("char[]", binascii.unhexlify(key))

            ### ABI leve, in-line call
            self.caesar.crypto_aead_encrypt(c, clen, m, mlen, ad, adlen, nsec, npub, key)

            ### Output
            # Note: Cannot use ffi.string() as it will truncate at null (0x00)
            output = "".join(binascii.hexlify(c[i]).decode("utf-8").upper() for i in range(0, clen[0]))

            encrypted     = output[:int((clen[0]-self.param['sizeTag'])*2)]
            sizeEncrypted = clen[0]-self.param['sizeTag']
            tag           = output[int((clen[0]-self.param['sizeTag'])*2):]
            # nsec_ciph     = binascii.hexlify(ffi.string(nsec,self.param['sizeNsec'])).decode("utf-8").upper()

            ### Decryption
            if (self.param['decrypt'] == True):
                print("\n\t===========================")
                print("\t        Decrypt")
                print("\t===========================\n")

                mlen    = ffi.new("unsigned long long *", 1)
                clen    = ffi.cast("unsigned long long", clen[0])
                self.caesar.crypto_aead_decrypt( m, mlen, nsec, c, clen, ad, adlen, npub, key)

        elif algorithm.lower() in ['keccak-hash', 'sha3']:
            alg = Keccak.Keccak()
            msg = (sizeMsg*8, msg)
            (tag, _) = alg.Hash(msg, n=256, r=1088, c=512,verbose=False)
        elif 'keccak-mac'       == algorithm.lower():
            alg = Keccak.Keccak()
            msg = (sizeMsg*8, msg)
            key = (self.param['sizeKey']*8, key)
            (tag, _) = alg.MAC(key, msg, n=256, r=1088, c=512,verbose=False)
        elif 'keccak-prng'      == algorithm.lower():
            alg = Keccak.Keccak()
            seed = (self.param['sizeNpub']*8, npub)
            (encrypted, _) = alg.PRNG(seed,n=512,r=1088,c=512,verbose=False)
        elif algorithm.lower() in ['lakekeyak', 'keccak-ae', 'keccak-keyak']:
            alg = Keccak.Keccak()
            M   = (sizeMsg*8, msg)
            N   = (self.param['sizeNpub']*8, npub)
            A   = (sizeAD*8, ad)
            K   = (self.param['sizeKey']*8, key)
            ([encrypted, tag], _) = alg.LakeKeyak(K, N, A, M,  n=128,r=1348,c=252,enc=1,verbose=True)
        else:
            raise ValueError("Unsupported algorithm : {0}.".format(algorithm))

        return (encrypted, sizeEncrypted, tag)

    def _genCipherCoreTvCHeader(self):
        args = {'lib'                   : self.param['caesarLib'],
                'ptm'                   : self.param['plainTextMode'],
                'ctm'                   : self.param['cipherTextMode'],
                'padAd'                 : self.param['padAD'],
                'padD'                  : self.param['padD'],
                'padStyle'              : self.param['padStyle'],
                'sizeElement'           : self.param['sizeElement'],
                'latencyEncTagCycles'   : self.param['Latency']['EncTagCycles'],
                'latencyDecTagCycles'   : self.param['Latency']['DecTagCycles'],
                'latencyNoMsgTagCyles'  : self.param['Latency']['NoMsgTagCycles'],
                }
        header = """/*
* This file was automatically generated by AETVgen.py
* for {lib} with the following implementation parameters:
* plainTextMode  : {ptm}
* CipherTextMode : {ctm}
* PAD_AD         : {padAd}
* PAD_D          : {padD}
* PAD_STYLE      : {padStyle}
* Element Size   : {sizeElement}
* Latency
*   EncTagCycles   : {latencyEncTagCycles}
*   DecTagCycles   : {latencyDecTagCycles}
*   NoMsgTagCycles : {latencyNoMsgTagCyles}
*/

#include <string.h>
#include <stdio.h>
#include "ap_cint.h"
#include "hls_tb.h"

int test(int test)
{{
\tunsigned char *data                 = calloc(200, sizeof(unsigned char));
\tuint{sizeElement} *bdi              = calloc(RateSize, sizeof(uint{sizeElement}));
\tuint{sizeElement} *bdo              = calloc(RateSize, sizeof(uint{sizeElement}));
\tuint{sizeElement} *key              = calloc(KeySize,  sizeof(uint{sizeElement}));
\tuint{sizeElement} *tag              = calloc(TagSize,  sizeof(uint{sizeElement}));
\tuint{sizeElement} *exp_tag          = calloc(TagSize,  sizeof(uint{sizeElement}));
\tuint{sizeElement} *npub             = calloc(NpubSize, sizeof(uint{sizeElement}));
\tuint{sizeElement} *nsec             = calloc(NsecSize, sizeof(uint{sizeElement}));
\tuint{sizeElement} *bdi_valid_bits   = calloc(RateSize, sizeof(uint{sizeElement}));
\tuint{sizeElement} *bdi_pad_loc      = calloc(RateSize, sizeof(uint{sizeElement}));
\tBDISIZE_t bdi_size      = 0;
\tBDOSIZE_t *bdo_size     = calloc(1, sizeof(BDOSIZE_t));
\tuint1 *msg_auth_valid   = calloc(1, sizeof(uint1));
\tuint1 *key_needs_update = calloc(1, sizeof(uint1)) ;
\tuint1 *force_start      = calloc(1, sizeof(uint1));
\tuint1 bdi_eoi = 0, bdi_eot = 1, bdi_ad = 1, bdi_decrypt = 0, bdi_nodata = 0;

\t*key_needs_update = 1;
\tmemset(npub, 0xB0, NpubSize);
\tmemset(key,  0x55, KeySize);
\tmemset(bdi,  0x00, RateSize);""".format(**args)

        #print(header.format(self.param['caesarLib']))
        self._pp(header, self.FILE_C)

    def _checkCInputsAndRun(self):
        self._pp("#if DISPLAY_INPUTS", self.FILE_C)
        self._pp("\t\tprintf(\"BDI      : \"); printData(bdi, RateSize);", self.FILE_C)
        self._pp("\t\tprintf(\"BDI vbits: \"); printData(bdi_valid_bits, RateSize);", self.FILE_C)
        self._pp("\t\tprintf(\"BDI ploc : \"); printData(bdi_pad_loc, RateSize);", self.FILE_C)
        self._pp("#endif", self.FILE_C)
        self._pp("\t\tCRYPTO_CALL;", self.FILE_C)

    def _genCipherCoreTvC(self, npub, nsec,
            paddedAd, sizeAd,
            paddedDin,  sizeDin,
            paddedDout, sizeDout,
            tag, newKey, decrypt):

        if self.param['plainTextMode'] in [1, 2]:
            paddedAd = npub + paddedAd if self.param['plainTextMode'] == 1 else paddedAd + npub
            sizeAd   = sizeAd + self.param['sizeNpub']
        
        rateSize = self.blockSize
        (decryptMsg, sizeMsg) = ("ENC", sizeDin) if decrypt == False else ("DEC", sizeDout)
        self._pp("\tif (test == {})".format(self.msgid), self.FILE_C)
        self._pp("\t{", self.FILE_C)
        self._pp("\t\tprintf(\"=============\\n\");", self.FILE_C)
        self._pp("\t\tprintf(\"  [{}] AD = {} byte, Msg = {} byte  \\n\");".format(decryptMsg, sizeAd, sizeMsg), self.FILE_C)
        self._pp("\t\tprintf(\"=============\\n\");", self.FILE_C)
        self._pp("\t\tstrcpy(data, \"{}\\0\");".format(tag), self.FILE_C)
        self._pp("\t\tstr2input(data, exp_tag, TagSize);", self.FILE_C)
        self._pp("\t\tstrcpy(data, \"{}\\0\");".format(self.key), self.FILE_C)
        self._pp("\t\tstr2input(data, key, KeySize);", self.FILE_C)
        self._pp("\t\tstrcpy(data, \"{}\\0\");".format(npub), self.FILE_C)
        self._pp("\t\tstr2input(data, npub, NpubSize);", self.FILE_C)
        if self.param['enableNsec'] == True:
            self._pp("\t\tstrcpy(data, \"{}\\0\");".format(nsec), self.FILE_C)
            self._pp("\t\tstr2input(data, nsec, NsecSize);", self.FILE_C)

        self._pp("#if DISPLAY_INPUTS", self.FILE_C)
        self._pp("\t\tprintf(\"Key    : \"); printData(key, KeySize);", self.FILE_C)
        self._pp("\t\tprintf(\"Npub   : \"); printData(npub, NpubSize);", self.FILE_C)
        self._pp("\t\tprintf(\"ExpTag : \"); printData(exp_tag, TagSize);", self.FILE_C)
        if self.param['enableNsec'] == True:
            self._pp("\t\tprintf(\"Nsec   : \"); printData(nsec, NsecSize);", self.FILE_C)
        self._pp("#endif", self.FILE_C)

        ### Initialization
        ccount = 1
        (isSize, isAd) = (0, 1)
        (isEoi, isEot, isNodata) = (0, 0, 0)
        self._pp("\n\t\tprintf(\"==(Call {} - Initialization)==\\n\");".format(ccount), self.FILE_C)
        self._pp("\t\t*key_needs_update = {};".format(1), self.FILE_C); # Note: Always initialize
        self._pp("\t\tbdi_decrypt = {};".format(int(decrypt)), self.FILE_C)
        self._pp("\t\tbdi_size = {}; bdi_ad = {};".format(isSize, isAd), self.FILE_C)
        self._pp("\t\tbdi_eoi = {}; bdi_eot = {}; bdi_nodata = {};".format(isEoi, isEot, isNodata), self.FILE_C)
        self._pp("\t\tCRYPTO_CALL;", self.FILE_C)
        ccount = ccount+1

        ### AD
        isAd = 1;
        totalBlocks = int((len(paddedAd)/2)/rateSize)
        for i in range(0,totalBlocks):
            isNodata = 0
            if (i+1) == totalBlocks:
                isSize = sizeAd % rateSize
                isEot  = 1
                isEoi  = 0 if len(paddedDin) > 0 else 1
            else:
                isSize = 0
                isEot  = 0
                isEoi  = 0
            self._pp("\n\t\tprintf(\"==(Call {} - AD[{}])==\\n\");".format(ccount,i), self.FILE_C)
            self._pp("\t\tmemset(bdi_valid_bits, 0x00, RateSize);",self.FILE_C)
            self._pp("\t\tmemset(bdi_pad_loc, 0x00, RateSize);",self.FILE_C)
            self._pp("\t\tbdi_size = {}; bdi_ad = {};".format(isSize, isAd), self.FILE_C)
            self._pp("\t\tbdi_eoi = {}; bdi_eot = {}; bdi_nodata = {};".format(isEoi, isEot, isNodata), self.FILE_C)
            self._pp("\t\tstrcpy(data, \"{}\\0\");".format(paddedAd[int(rateSize*2*i):int(rateSize*2*(i+1))]), self.FILE_C)
            self._pp("\t\tstr2input(data, bdi, RateSize);", self.FILE_C)
            if isSize == 0:
                self._pp("\t\tmemset(bdi_valid_bits, 0xFF, RateSize);",self.FILE_C)
            else:
                self._pp("\t\tmemset(bdi_valid_bits, 0xFF, {});".format(isSize),self.FILE_C)
                self._pp("\t\tmemset(bdi_pad_loc+{}, 0x80, 1);".format(isSize),self.FILE_C)
            self._checkCInputsAndRun()
            ccount = ccount + 1

        ### Input & Expected output
        isAd = 0;
        totalBlocks = int((len(paddedDin)/2)/rateSize)
        for i in range(0,totalBlocks):
            self._pp("\n\t\tprintf(\"====(Call {} - Data[{}])====\\n\");".format(ccount,i), self.FILE_C)
            if totalBlocks == (i+1):
                isSize = sizeDin % rateSize
                isEot  = 1
                isEoi  = 1
                isNodata = 1 if (int(len(paddedDin)/2)-rateSize) == sizeDin else 0
            else:
                isSize = 0
                isEot  = 0
                isEoi  = 0
                isNodata = 0

            # Handle PRIMATEs-APE (size)
            if self.param['cipherTextMode'] == 2 and self.param['reverseDblk'] == True and decrypt == True:
                isSize = sizeDin % rateSize if i == 0 else 0
                    

            self._pp("\t\tmemset(bdi_valid_bits, 0x00, RateSize);",self.FILE_C)
            self._pp("\t\tmemset(bdi_pad_loc, 0x00, RateSize);",self.FILE_C)
            self._pp("\t\tbdi_size = {}; bdi_ad = {};".format(isSize, isAd), self.FILE_C)
            self._pp("\t\tbdi_eoi = {}; bdi_eot = {}; bdi_nodata = {};".format(isEoi, isEot, isNodata), self.FILE_C)
            # Input
            self._pp("\t\tstrcpy(data, \"{}\\0\");".format(paddedDin[int(rateSize*2*i):int(rateSize*2*(i+1))]), self.FILE_C)
            self._pp("\t\tstr2input(data, bdi, RateSize);", self.FILE_C)
            if isSize == 0:
                if isNodata == 0:
                    self._pp("\t\tmemset(bdi_valid_bits, 0xFF, RateSize);", self.FILE_C)
            else:
                self._pp("\t\tmemset(bdi_valid_bits, 0xFF, {});".format(isSize), self.FILE_C)
                self._pp("\t\tmemset(bdi_pad_loc+{}, 0x80, 1);".format(isSize), self.FILE_C)
            self._checkCInputsAndRun()
            ccount = ccount + 1
            # Expected output
            exp_out = paddedDout[int(rateSize*2*i):int(rateSize*2*(i+1))]
            if self.param['cipherTextMode'] == 2:
                self._pp("\t\tprintf(\"BDOsize: %d\\n\", *bdo_size);", self.FILE_C)
            self._pp("\t\tprintf(\"BDO    : \"); printData(bdo, RateSize);", self.FILE_C)
            self._pp("\t\tprintf(\"Exp Out: {}\\n\");".format(exp_out), self.FILE_C)
            if len(exp_out) > 0:
                self._pp("\t\tif (compareData(\"{}\", bdo, {}, {}))".format(exp_out, rateSize, int(decrypt)), self.FILE_C)
                self._pp("\t\t\treturn(1);", self.FILE_C);

        # Tag
        if decrypt == False:
            cycles = self.param['Latency']['NoMsgTagCycles'] if (len(paddedDin) == 0) else self.param['Latency']['EncTagCycles']
            for i in range(0, cycles):
                self._pp("\n\t\tprintf(\"====(Call {} - Tag[{}])====\\n\");".format(ccount,i), self.FILE_C)
                self._pp("\t\tif (*force_start == 1) CRYPTO_CALL; else printf(\"Not started.\\n\");", self.FILE_C)
                ccount = ccount + 1
                cycles = cycles - 1
            self._pp("\t\tprintf(\"Out Tag: \"); printData(tag, TagSize);", self.FILE_C)
            self._pp("\t\tprintf(\"Exp Tag: {}\\n\");".format(tag), self.FILE_C)
            self._pp("\t\tif (compareData(\"{}\", tag, {}))".format(tag, self.param['sizeTag'], int(decrypt)), self.FILE_C)
            self._pp("\t\t\treturn(1);", self.FILE_C);
        else:
            cycles = self.param['Latency']['NoMsgTagCycles'] if (len(paddedDin) == 0) else self.param['Latency']['DecTagCycles']
            for i in range(0, cycles):
                self._pp("\t\t\nprintf(\"====(Call {} - Tag[{}])====\\n\");".format(ccount,i), self.FILE_C)
                self._pp("\t\tif (*force_start == 1) CRYPTO_CALL; else printf(\"Not started.\\n\");", self.FILE_C)
                ccount = ccount + 1
                cycles = cycles - 1
            self._pp("\t\tprintf(\"TagVlid: %d\\n\", *msg_auth_valid);", self.FILE_C)

        # End
        txt = "!(*msg_auth_valid)" if decrypt else "0"
        self._pp("\t\treturn({});".format(txt), self.FILE_C)
        self._pp("\t}", self.FILE_C)

    def genAeadTv(self, algorithm, opcode, npub, sizeNpub,
              ad, sizeAD, msg, sizeMsg,
              encrypted, sizeEncrypted,
              nsec, tag, newKey):
        """ Generate a single test vector in VHDL for AEAD """
        self.msgid = self.msgid + 1
        sizeInput  = sizeMsg
        sizeOutput = sizeEncrypted

        ### Message header information
        text = "Ciphertext" if opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT] else "Plaintext"
        self._pp("\n#### MsgID={0:03}, KeyID={1:03}, AD Size={2: 4}, {6} Size={3: 4}, Opcode={4}, Algorithm={5}".format(self.msgid, self.keyid, sizeAD, sizeMsg, self.CODE_OPCODE[opcode], algorithm, text), self.FILE_PUBLIC)

        (pad, eoi, eot) = (0, 0, 0)

        ### Generate New Key, if required.
        if (newKey == 1 or self.key == '') and \
           (opcode in \
                [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, \
                 self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, \
                 self.OPCODE_MAC, self.OPCODE_PRNG]):
            ## SDI (Key segment)
            (pad, eoi, eot) = (0, 1, 1)
            # Instruction
            self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, self.CODE_OPCODE[self.OPCODE_LOAD_KEY]), self.FILE_SECRET)
            self._pp("INS = {0}".format(self.getHdrIns(self.OPCODE_LOAD_KEY,self.param['sizeSIO'])), self.FILE_SECRET)
            # Header + data
            if (self.param['enableRoundKey'] == True):
                totalKeySize = self.param['sizeRoundKey']*self.param['totalRoundKey']
                stype = self.STYPE_ROUNDKEY
                data  = 'X'*(totalKeySize*2)
            else:
                totalKeySize = self.param['sizeKey']
                stype = self.STYPE_KEY
                data  = self.key
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[stype], totalKeySize, pad, eoi, eot), self.FILE_SECRET)
            if (self.param['enableRoundKey'] == True):
                self._pp("# Round key generated from key : {}".format(self.key), self.FILE_SECRET)
                self._pp("# Note: Replace XXX... with actual round key", self.FILE_SECRET)
            self._pp("HDR = {0}".format(self.getHdrCmd(stype, pad, eoi, eot, totalKeySize, self.param['sizeSIO'])), self.FILE_SECRET)
            self._pp("DAT = {0}{1}".format(data, '0'*self._calcIOFill(totalKeySize*2,self.param['sizeSIO'])), self.FILE_SECRET)

            ## PDI, DO (Key activation instruction)
            self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, self.CODE_OPCODE[self.OPCODE_ACT_KEY]), self.FILE_PUBLIC)
            self._pp("INS = {0}".format(self.getHdrIns(self.OPCODE_ACT_KEY,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            if opcode in [self.OPCODE_ENCRYPT, self.OPCODE_AEAD_ENCRYPT]:
                self._pp("INS = {0}".format(self.getHdrIns(self.OPCODE_ACT_KEY,self.param['sizePIO'])), self.FILE_OUTPUT)
        else:
            if opcode in [self.OPCODE_ENCRYPT, self.OPCODE_AEAD_ENCRYPT]:
                self._pp("# {0:03} : Instruction (Opcode={1} [Generated])".format(self.msgid, self.CODE_OPCODE[self.OPCODE_ACT_KEY]), self.FILE_OUTPUT)
                self._pp("INS = {0}".format(self.getHdrIns(self.OPCODE_ACT_KEY,self.param['sizePIO'])), self.FILE_OUTPUT)
            elif opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT]:
                self._pp("# {0:03} : Instruction (Opcode={1} [Same key])".format(self.msgid, self.CODE_OPCODE[self.OPCODE_ACT_KEY]), self.FILE_PUBLIC_INPUT)
                self._pp("INS = {0}".format(self.getHdrIns(self.OPCODE_ACT_KEY,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)

        if self.param['enableNsec'] == True:
            nsec_ciph = encrypted[:self.param['sizeNsec']*2]
            encrypted = encrypted[self.param['sizeNsec']*2:]
            sizeOutput = sizeOutput - self.param['sizeNsec']
        else:
            nsec_ciph = "0"

        ### Message handling for APE
        if ((self.param['cipherTextMode'] == 2) and (self.param['reverseDblk'] == True)):
            ### Create new output with appropriate format
            # print(encrypted)

            newEncrypted = ""
            fullBlocks = int(len(encrypted)/(self.blockSize*2))
            partBlock  = len(encrypted) % (self.blockSize*2)
            if (fullBlocks < 2):
                newEncrypted = newEncrypted + self._xpad(encrypted[:(len(encrypted)%(self.blockSize*2))],self.param['sizePIO'])
                for i in range(0, fullBlocks):
                    newEncrypted = newEncrypted + encrypted[(len(encrypted)%(self.blockSize*2))+i*self.blockSize*2:(len(encrypted)%(self.blockSize*2))+(i+1)*self.blockSize*2]
            else:
                newEncrypted = encrypted[:((fullBlocks-1)*self.blockSize*2)]
                newEncrypted = newEncrypted + self._xpad(encrypted[int((fullBlocks-1)*self.blockSize*2):int((fullBlocks-1)*self.blockSize*2)+partBlock],self.param['sizePIO'])
                newEncrypted = newEncrypted + encrypted[int((fullBlocks-1)*self.blockSize*2)+partBlock:]

            if opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT]:
                # Re-order input
                encrypted = ""
                for i in range(0, int(len(newEncrypted)/(self.blockSize*2))):
                    encrypted = encrypted + newEncrypted[int(len(newEncrypted)-(i+1)*self.blockSize*2):int(len(newEncrypted)-i*self.blockSize*2)]
                encrypted = encrypted.replace('X','0')

                newMsg = self._xpad(msg,self.param['sizePIO'])
                msg    = ""
                for i in range(0, int(len(newMsg)/(self.blockSize*2))):
                    msg = msg + newMsg[int(len(newMsg)-(i+1)*self.blockSize*2):int(len(newMsg)-i*self.blockSize*2)]
            else:
                encrypted = newEncrypted

        ### Swap msg and encrypted data if decryption
        if opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT]:
            (encrypted, msg)        = (msg, encrypted)
            (sizeInput, sizeOutput) = (sizeOutput, sizeInput)
            (nsec_ciph, nsec)       = (nsec, nsec_ciph)
            ### Error checking
            if sizeOutput != sizeInput:
                if self.param['cipherTextMode'] != 2:
                    print("input : " , sizeInput)
                    print("output: " , sizeOutput)
                    self._printCallInfo(algorithm, opcode, self.key, npub, ad, msg, encrypted, tag)
                    raise ValueError("Error! encrypted output length is not the same as the input message length.")

        ### Instruction
        # Input
        self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, self.CODE_OPCODE[opcode]), self.FILE_PUBLIC_INPUT)
        self._pp("INS = {0}".format(self.getHdrIns(opcode,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
        # Output
        if   opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_ENCRYPT]:
            opc_text = self.CODE_OPCODE[opcode+1]
            opc      = opcode + 1
        elif opcode in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]:
            opc_text = self.CODE_OPCODE[self.OPCODE_AE_PASS]
            opc      = self.OPCODE_AE_PASS
        self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, opc_text), self.FILE_OUTPUT)
        self._pp("INS = {0}".format(self.getHdrIns(opc,self.param['sizePIO'])), self.FILE_OUTPUT)

        ### Length segment
        if self.param['enableLoadLen'] == True:
            type    = self.STYPE_LEN            
            len_ad  = self._zpad("{size:0{width}X}".format(size=sizeAD, width=self.param['ctrSizeAd']*2), self.param['sizePIO'])
            len_d   = self._zpad("{size:0{width}X}".format(size=sizeMsg, width=self.param['ctrSizeD']*2), self.param['sizePIO'])
            # Input
            (pad, eoi, eot) = (0, 0, 1)
            size = int((len(len_ad) + len(len_d))/2)
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[type],size,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(type,pad,eoi,eot,size,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("DAT = {0}".format(len_ad + len_d), self.FILE_PUBLIC_INPUT)
            # Output (not exist if Decrypt)
            if opcode not in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]:
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[type],size,pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(type,pad,eoi,eot,size,self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(len_ad + len_d), self.FILE_OUTPUT)

        ### Npub
        if (sizeNpub > 0) and (self.param['plainTextMode'] == 0) and \
           (opcode in [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, \
                       self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, \
                       self.OPCODE_PRNG]):
            eot = 1
            eoi = 1 if (sizeInput == 0) and (sizeAD == 0) else 0
            # Input PDI file
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_NPUB],self.param['sizeNpub'],pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_NPUB,pad,eoi,eot,self.param['sizeNpub'],self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("DAT = {0}".format(self._zpad(npub,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            # Output DO file (remove if AEAD_DECRYPT)
            if opcode not in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]:
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_NPUB],self.param['sizeNpub'],pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_NPUB,pad,eoi,eot,self.param['sizeNpub'],self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(self._zpad(npub,self.param['sizePIO'])), self.FILE_OUTPUT)
        elif (self.param['plainTextMode'] == 1): # NA_M
            self._pp("# Npub is prepended to AD", self.FILE_PUBLIC_INPUT)
            self._pp("# Npub = {0}".format(npub), self.FILE_PUBLIC)
            ad = npub + ad
            sizeAD = sizeAD + self.param['sizeNpub']
        elif (self.param['plainTextMode'] == 2): # AN_M
            self._pp("# Npub is appended to AD", self.FILE_PUBLIC_INPUT)
            self._pp("# Npub = {0}".format(npub), self.FILE_PUBLIC)
            ad = ad + npub
            sizeAD = sizeAD + self.param['sizeNpub']

        ### Nsec
        if (self.param['enableNsec'] == True) and \
           (opcode in [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT]):
            if opcode == self.OPCODE_DECRYPT or opcode == self.OPCODE_AEAD_DECRYPT:
                code_nsec_in  = self.STYPE_ENC_NSEC
                code_nsec_out = self.STYPE_NSEC
            else:
                code_nsec_in  = self.STYPE_NSEC
                code_nsec_out = self.STYPE_ENC_NSEC
            (eot, eoi) = (1, 0)
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[code_nsec_in], self.param['sizeNsec'],pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(code_nsec_in,pad,eoi,eot,self.param['sizeNsec'],self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("DAT = {0}".format(self._zpad(nsec,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[code_nsec_out],self.param['sizeNsec'],pad,eoi,eot), self.FILE_OUTPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(code_nsec_out,pad,eoi,eot,self.param['sizeNsec'],self.param['sizePIO'])), self.FILE_OUTPUT)
            self._pp("DAT = {0}".format(self._xpad(nsec_ciph,self.param['sizePIO'])), self.FILE_OUTPUT)


        if ((opcode == self.OPCODE_AEAD_DECRYPT) and (self.param['reverseDblk'] == True) and (self.param['cipherTextMode'] == 2)):
            (pad, eoi, eot) = (0, 0, 1)
            self._genAeadTagSegment(tag, opcode, pad, eoi, eot)

        ### AD
        if opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT]:
            (eoi, eot)      = (0, 0)
            maxSizeSegment  = self.param['maxSizeSegment']
            fullSegment     = int(sizeAD/maxSizeSegment)
            finalSgtSize    = sizeAD % maxSizeSegment
            # Full blocks
            for i in range(0,fullSegment):
                if i == fullSegment-1 and finalSgtSize == 0:
                    eot = 1
                    if ((sizeInput > 0 and opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_ENCRYPT]) or
                       (sizeOutput > 0 and opcode in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT])):
                        self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],maxSizeSegment,pad,eoi,eot), self.FILE_PUBLIC)
                        self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,maxSizeSegment,self.param['sizePIO'])), self.FILE_PUBLIC)
                    else:
                         # AD (input)
                        eoi = 0 if (self.param['cipherTextMode'] == 2 and self.param['padD'] == 4 and opcode in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]) else 1
                        self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],maxSizeSegment,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                        self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,maxSizeSegment,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                        # AD (output)
                        eoi = 0 if (self.param['cipherTextMode'] == 2 and self.param['padD'] == 4 and opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_ENCRYPT]) else 1
                        self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],maxSizeSegment,pad,eoi,eot), self.FILE_SECRET)
                        self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,maxSizeSegment,self.param['sizePIO'])), self.FILE_OUTPUT)
                    self._pp("DAT = {0}".format(ad[i*maxSizeSegment*2:(i+1)*maxSizeSegment*2]), self.FILE_PUBLIC)
                else:
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],maxSizeSegment,pad,eoi,eot), self.FILE_PUBLIC)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,maxSizeSegment,self.param['sizePIO'])), self.FILE_PUBLIC)
                    self._pp("DAT = {0}".format(ad[i*maxSizeSegment*2:(i+1)*maxSizeSegment*2]), self.FILE_PUBLIC)
            #   Partial block
            if finalSgtSize > 0:
                eot = 1
                if ((sizeInput  > 0 and opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_ENCRYPT]) or
                   (sizeOutput > 0 and opcode in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT])):
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],finalSgtSize,pad,eoi,eot), self.FILE_PUBLIC)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_PUBLIC)
                else:
                    # AD (input)
                    eoi = 0 if (self.param['cipherTextMode'] == 2 and self.param['padD'] == 4 and opcode in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]) else 1
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],finalSgtSize,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                    # AD (output)
                    eoi = 0 if (self.param['cipherTextMode'] == 2 and self.param['padD'] == 4 and opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_ENCRYPT]) else 1
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],finalSgtSize,pad,eoi,eot), self.FILE_OUTPUT)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(self._zpad(ad[int(fullSegment*maxSizeSegment*2):],self.param['sizePIO'])), self.FILE_PUBLIC)


        ### Data (Input)
        if (sizeInput > 0) and (opcode not in [self.OPCODE_PRNG]):
            if opcode == self.OPCODE_DECRYPT or opcode == self.OPCODE_AEAD_DECRYPT:
                itype = self.STYPE_CIPHERTEXT
            else:
                itype = self.STYPE_PLAINTEXT
            (eoi, eot)      = (0, 0)
            maxSizeSegment  = self.param['maxSizeSegment']
            if (self.param['cipherTextMode'] == 2 and self.param['padD'] == 4 and opcode in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT] and sizeOutput % self.param['maxSizeSegment'] == 0):   # Prevent a segment to contain only a padded block
                fullSegment     = 0 if (sizeInput-maxSizeSegment < 0) else int((sizeInput-maxSizeSegment)/self.param['maxSizeSegment'])
            else:
                fullSegment     = int(sizeInput/maxSizeSegment)
            finalSgtSize    = sizeInput - maxSizeSegment*fullSegment
            # Full blocks
            for i in range(0,fullSegment):
                if i == fullSegment-1 and finalSgtSize == 0:
                    (eot, eoi) = (1, 1)
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[itype],maxSizeSegment,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(itype,pad,eoi,eot,maxSizeSegment,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                self._pp("DAT = {0}".format(msg[int(i*maxSizeSegment*2):int((i+1)*maxSizeSegment*2)]), self.FILE_PUBLIC_INPUT)

            # Partial blocks
            if finalSgtSize > 0:
                (eot, eoi) = (1, 1)
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[itype],finalSgtSize,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(itype,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                self._pp("DAT = {0}".format(self._zpad(msg[int(fullSegment*maxSizeSegment*2):],self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
        elif (opcode in [self.OPCODE_PRNG]):
            itype = self.STYPE_PLAINTEXT
            (eot, eoi) = (1, 1)
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[itype],sizeInput,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(itype,pad,eoi,eot,sizeInput,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)

        ### Data (Output)
        if  (sizeOutput > 0) and \
            (opcode in [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, \
                        self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, \
                        self.OPCODE_PRNG]):
            if opcode in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]:
                otype = self.STYPE_PLAINTEXT
            else:
                otype = self.STYPE_CIPHERTEXT
            (eoi, eot)      = (0, 0)
            maxSizeSegment  = self.param['maxSizeSegment']
            if (self.param['cipherTextMode'] == 2 and self.param['padD'] == 4 and opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_ENCRYPT] and sizeInput % self.param['maxSizeSegment'] == 0):   # Prevent a segment to contain only a padded block
                fullSegment     = 0 if ((sizeOutput-maxSizeSegment) < 0) else int((sizeOutput-maxSizeSegment)/self.param['maxSizeSegment'])
            else:
                fullSegment     = int(sizeOutput/maxSizeSegment)
            finalSgtSize    = sizeOutput - maxSizeSegment*fullSegment
            # Full blocks
            for i in range(0,fullSegment):
                if i == fullSegment-1 and finalSgtSize == 0:
                    (eot, eoi) = (1, 1)
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[otype],maxSizeSegment,pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(otype,pad,eoi,eot,maxSizeSegment,self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(encrypted[int(i*maxSizeSegment*2):int((i+1)*maxSizeSegment*2)]), self.FILE_OUTPUT)

            # Partial blocks
            if finalSgtSize > 0:
                (eot, eoi) = (1, 1)
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[otype],finalSgtSize,pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(otype,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(self._xpad(encrypted[int(fullSegment*maxSizeSegment*2):],self.param['sizePIO'])), self.FILE_OUTPUT)


        ### Tag
        if opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, self.OPCODE_MAC, self.OPCODE_HASH]:
            if not ((opcode == self.OPCODE_AEAD_DECRYPT) and (self.param['reverseDblk'] == True) and (self.param['cipherTextMode'] == 2)):
                eoi = 1
                self._genAeadTagSegment(tag, opcode, pad, eoi, eot)
                
    
    def getInputData(self, newKey, sizeAD, sizeMsg):
        nsec = "00"
        msg  = ""
        if self.param['testMode'] >= 1:
            npub   = "B0"*self.param['sizeNpub']
            if self.param['testMode'] == 1:
                msg  = "FF"*sizeMsg
                ad   = "A0"*sizeAD
            else:
                msg = "".join("{0:02X}".format(j%256) for j in range(1,sizeMsg+1))
                ad  = "".join("{0:02X}".format(j%256) for j in range(1,sizeAD+1))
            if self.param['enableNsec'] == True:
                nsec = "66"*self.param['sizeNsec']
        else:
            if (self.param['sizeNpub'] > 0):
                npub = self._genRandomData(self.param['sizeNpub'])
            else:
                npub = ""
            ad   = self._genRandomData(sizeAD)
            msg  = self._genRandomData(sizeMsg)
            if self.param['enableNsec'] == True:
                nsec = self._genRandomData(self.param['sizeNsec'])
        if (newKey == 1 or self.key == ''):
            self.keyid += 1
            newKey     = 1
            if self.param['testMode'] >= 1:
                self.key = "55"*self.param['sizeKey']
            else:
                self.key = self._genRandomData(self.param['sizeKey'])
        return (npub, ad, msg, nsec)
        
    ### Single 
    def genEncDecTestVector(self, sizeAD, sizeMsg):
        opcode = random.getrandbits(1)+2
        newKey = 1
        
        ### Print header
        self._printInfo()        
        print("Generating message {}/{} {} [AD Sizes={:< 2},Msg Size={}]".format(1,1,self.CODE_OPCODE[opcode],int(sizeAD),int(sizeMsg)))

        ### Generate input data
        (npub, ad, msg, nsec) = self.getInputData(newKey, sizeAD, sizeMsg)

        ### Compute output
        (encrypted, sizeEncrypted, tag) = self.compute(self.param['algorithm'], opcode, self.key, npub, ad, sizeAD, msg, sizeMsg, nsec)

        ### Generate AEAD TV
        self.genAeadTv(self.param['algorithm'], opcode,
                   npub, self.param['sizeNpub'],
                   ad, int(sizeAD),
                   msg, int(sizeMsg),
                   encrypted, sizeEncrypted,
                   nsec, tag, newKey)

        ### Generate Cipher Core TV
        if (self.param['genCipherCoreTvC'] or self.param['genCipherCoreTvVhdl']) == True:
            decrypt     = True             if opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT] else False
            (din, sizeDin, dout, sizeDout) = (msg, sizeMsg, encrypted, sizeEncrypted) if decrypt == False else (encrypted, sizeEncrypted, msg, sizeMsg)                
            (paddedAd, paddedDin, paddedDout) = self.pad(decrypt, ad, sizeAD, din, sizeDin, dout, sizeDout)
            
            if sizeDout == 0:
                dout = "XX"
            
            self._genCipherCoreTvC(npub, nsec,
                paddedAd, sizeAD,
                paddedDin,  sizeDin,
                paddedDout, sizeDout,
                tag, newKey, decrypt)
        self._pp("###EOF", self.FILE_ALL)
        if self.param['stdout'] == False:
            self._fileIO('close')
        print("Terminating AETVgen.py")                

        
    ### Main calling function
    def genTestVectors(self, sampleSize):
        """ Generate Test Vectors """
        # Print settings
        self._printInfo()
        random.seed()
        if (self.param['testMode'] == 0):
            self.param['startTV'] = 0
        for i in range(self.param['startTV'],sampleSize):
            newKey  = random.getrandbits(1)
            if self.param['opcode'] != False:
                if len(self.param['opcode']) > 1:
                    opcode = self.param['opcode'][random.randrange(len(self.param['opcode']))]
                else:
                    opcode = self.param['opcode']
            else:
                opcode = random.getrandbits(1)+2 # Make sure that the opcode is always AEAD for now

            if (self.param['testMode'] == 0):
                sizeAD  = random.randrange(self.param['minSizeAD'], self.param['maxSizeAD']+1, 1)
                sizeMsg = random.randrange(self.param['minSize'],  self.param['maxSize']+1, 1)
                # Prevent AD = 0 and Msg = 0 case
                if (sizeAD == 0 and sizeMsg == 0):
                    sizeMsg = 1
            else:
                if (i % 2 == 1):
                    opcode = 3
                else:
                    opcode = 2
                if i in [0, 1]:
                    sizeMsg = 0
                    sizeAD  = 1
                elif i in [2, 3]:
                    sizeMsg = 1
                    sizeAD  = 0
                elif i in [4, 5]:
                    sizeMsg = 1
                    sizeAD  = 1
                elif i in [6, 7]:
                    sizeMsg = self.blockSize
                    sizeAD  = self.blockSizeAD
                elif i in [8, 9]:
                    sizeMsg = self.blockSize     - 1
                    sizeAD  = self.blockSizeAD   - 1
                elif i in [10, 11]:
                    sizeMsg = self.blockSize     + 1
                    sizeAD  = self.blockSizeAD   + 1
                elif i in [12, 13]:
                    sizeMsg = 2*self.blockSize
                    sizeAD  = 2*self.blockSizeAD
                elif i in [14, 15]:
                    sizeMsg = int(2*self.blockSize   + self.blockSize/2)
                    sizeAD  = int(2*self.blockSizeAD + self.blockSizeAD/2)
                elif i in [16, 17]:
                    sizeMsg = random.randrange(self.blockSize, 3*self.blockSize, 1)
                    sizeAD  = sizeMsg
                    while sizeAD == sizeMsg:
                        sizeAD = random.randrange(self.blockSize, 3*self.blockSize, 1)
                else:
                    sizeMsg = (int((i-12)/2)+1)*self.blockSize-1
                    sizeAD  = int(2*self.blockSizeAD-1+ (i/2 % 3))
            print("Generating message {}/{} {} [AD Sizes={:< 2},Msg Size={}]".format(i+1,sampleSize,self.CODE_OPCODE[opcode],int(sizeAD),int(sizeMsg)))

            ### Generate input data
            (npub, ad, msg, nsec) = self.getInputData(newKey, sizeAD, sizeMsg)

            ### Compute output
            (encrypted, sizeEncrypted, tag) = self.compute(self.param['algorithm'], opcode, self.key, npub, ad, sizeAD, msg, sizeMsg, nsec)

            ### Generate AEAD TV
            self.genAeadTv(self.param['algorithm'], opcode,
                       npub, self.param['sizeNpub'],
                       ad, int(sizeAD),
                       msg, int(sizeMsg),
                       encrypted, sizeEncrypted,
                       nsec, tag, newKey)

            ### Generate Cipher Core TV
            if (self.param['genCipherCoreTvC'] or self.param['genCipherCoreTvVhdl']) == True:
                decrypt     = True             if opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT] else False
                (din, sizeDin, dout, sizeDout) = (msg, sizeMsg, encrypted, sizeEncrypted) if decrypt == False else (encrypted, sizeEncrypted, msg, sizeMsg)                
                (paddedAd, paddedDin, paddedDout) = self.pad(decrypt, ad, sizeAD, din, sizeDin, dout, sizeDout)
                
                if sizeDout == 0:
                    dout = "XX"
                
                self._genCipherCoreTvC(npub, nsec,
                    paddedAd, sizeAD,
                    paddedDin,  sizeDin,
                    paddedDout, sizeDout,
                    tag, newKey, decrypt)

        self._pp("###EOF", self.FILE_ALL)
        if self.param['stdout'] == False:
            self._fileIO('close')
        print("Terminating AETVgen.py")

    def pad(self, decrypt, ad, sizeAd, din, sizeDin, dout, sizeDout):
        if self.param['padStyle'] == 1:
            padStyle = "80"
        else:
            padStyle = "00"

        ret = []        
        # Handle PRIMATEs-APE
        if self.param['cipherTextMode'] == 2 and self.param['reverseDblk'] == True:
            # Ad
            if (sizeAd % self.blockSize) > 0:
                ret.append(self._zpad(ad + padStyle, self.blockSize))
            else:
                ret.append(ad)
            totalBlocks = sizeDin/self.blockSize
            
            # Data            
            lastBlock = sizeDin % self.blockSize 
            if not decrypt:
                # Encrypt
                ret.append(self._zpad(din, self.blockSize))                
                if lastBlock > 0 and sizeDin > self.blockSize:
                    ret.append(self._xpad(dout[:sizeDin*2-self.blockSize*2],self.blockSize) + dout[sizeDin*2-self.blockSize*2:])
                else:
                    ret.append(self._xpad(dout, self.blockSize))
            else:        
                # Decrypt                
                if lastBlock > 0 and sizeDin > self.blockSize:
                    paddedDin  = self._zpad(din[:sizeDin*2-self.blockSize*2],self.blockSize) + din[sizeDin*2-self.blockSize*2:]
                    paddedDout = self._xpad(dout[:sizeDin*2-self.blockSize*2],self.blockSize) + dout[sizeDin*2-self.blockSize*2:]
                else:
                    paddedDin  = din
                    paddedDout = self._xpad(dout, self.blockSize)
                # Reverse order
                totalBlocks = int((len(paddedDin)/2)/self.blockSize)
                tmpi = ""
                tmpo = ""
                for i in range(0, totalBlocks):
                    tmpi += paddedDin[(totalBlocks-1-i)*self.blockSize*2:(totalBlocks-i)*self.blockSize*2]
                    tmpo += paddedDout[(totalBlocks-1-i)*self.blockSize*2:(totalBlocks-i)*self.blockSize*2]
                ret.append(tmpi)
                ret.append(tmpo)
        else:        
            for i, (padMode, data, size, bs) in \
                     enumerate(zip([self.param['padAD'], self.param['padD']],
                     [ad, din], [sizeAd, sizeDin],
                     [self.blockSizeAD, self.blockSize])):
                sizeLast = size%bs
                if padMode > 0:
                    # Add a padding byte
                    if sizeLast != 0:
                        data += padStyle
                    elif         size == 0  and padMode in [2, 4] or \
                        ((size % bs) == 0) and padMode in [3, 4]:
                        if not (i == 1 and decrypt == True and self.param['cipherTextMode'] == 2):
                            data +=  padStyle
                # Zero padded block
                data = self._zpad(data, bs)

                # Add to a list
                ret.append(data)
            # Dout
            ret.append(self._xpad(dout, self.blockSize))
        
        
        return ret

    ### =================
    ### Helper functions
    ### =================
    ### Generate Tag segment
    def _genAeadTagSegment(self, tag, opcode, pad, eoi, eot):
        # Input
        if opcode == self.OPCODE_AEAD_DECRYPT:
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_TAG],int(len(tag)/2),pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_TAG,pad,eoi,eot,len(tag)/2,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("DAT = {0}".format(self._xpad(tag,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
        # Output
        if opcode != self.OPCODE_AEAD_DECRYPT:  # No output TAG for decryption
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_TAG],int(len(tag)/2),pad,eoi,eot), self.FILE_OUTPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_TAG,pad,eoi,eot,len(tag)/2,self.param['sizePIO'])), self.FILE_OUTPUT)
            self._pp("DAT = {0}".format(self._xpad(tag,self.param['sizePIO'])), self.FILE_OUTPUT)

    def _xpad(self, text, iowidth):
        """ 'X' pad the text to match the I/O width """
        if self.param['dontCare'] == True:
            pad = 'X'
        else:
            pad = '0'
        return text + pad*((-len(text)) % (iowidth*2))

    def _zpad(self, text, iowidth):
        """ Zero pad the text to match the I/O width """
        return text + '0'*((-len(text)) % (iowidth*2))

    def _calcIOFill(self, filled, iowidth):
        return (iowidth*2-filled) % (iowidth*2)

    def _fileIO(self, op='open'):
        if op=='open':
            print ("Opening files...")
            self.pdi = open(self.param['filePDI'], "wt")
            self.sdi = open(self.param['fileSDI'], "wt")
            self.do  = open(self.param['fileDO'],  "wt")
            if self.param['genCipherCoreTvC'] == True:
                self.cfile = open(self.param['fileC'],  "wt")
                self._genCipherCoreTvCHeader()
            if self.param['genCipherCoreTvVhdl'] == True:
                self.hdl  = open(self.param['fileVhdl'], "wt")
        else:
            print ("Closing files...")
            self.pdi.close()
            self.sdi.close()
            self.do.close()
            if self.param['genCipherCoreTvC'] == True:
                self.cfile.write("\n } ")
                self.cfile.close()
            if self.param['genCipherCoreTvVhdl'] == True:
                self.hdl.close()

    def _genRandomData(self, size, base='HEX'):
        """ Generate random data """
        if (size == 0):
            return ("00")
        else:
            if base=='HEX':
                return ("%0*X" % (size*2, random.randrange(256**size)))
            else:
                print ("Error! Unsupported <base> for function _genRandomData(<size>, <base>)")
                print ("base = %s" % (base))

    def _printCallInfo(self, algorithm, opcode, key, npub, ad, msg, encrypted, tag):
        print ("\n\n")
        print ("{0} - {1}".format(algorithm, self.CODE_OPCODE[opcode]))
        print ("Key\t\t:{0}".format(key))
        print ("Npub\t\t:{0}".format(npub))
        print ("Ad\t\t:{0}".format(ad))
        print ("Msg\t\t:{0}".format(msg))
        print ("Enc(Msg)\t:{0}".format(encrypted))
        print ("Tag\t\t:{0}".format(tag))
        print ("Len(Msg)\t:{0}".format(int(len(msg)/2)))
        print ("Len(encrypted)\t:{0}".format(int(len(encrypted)/2)))
        print ("Len(tag)\t:{0}".format(int(len(tag)/2)))

    def _printInfo(self, additionalIgnores=[]):
        ignoreParams = ['stdout', 'filePDI', 'fileSDI', 'fileDO',
            'fileC', 'fileVhdl',
            'Latency', 'genCipherCoreTvC', 'genCipherCoreTvVhdl'] + \
            additionalIgnores
        if self.param['stdout'] == False:
            self._fileIO('open')

        for target in self.FILE_SECRET, self.FILE_PUBLIC_INPUT, self.FILE_OUTPUT:
            self._pp("# ===================================================", target)
            if self.param['stdout'] == False:
                if target == self.FILE_SECRET:
                    self._pp("# {0}".format(self.sdi.name),                   target)
                elif target == self.FILE_PUBLIC_INPUT:
                    self._pp("# {0}".format(self.pdi.name),                   target)
                elif target == self.FILE_OUTPUT:
                    self._pp("# {0}".format(self.do.name),                    target)
            self._pp("# This file was generated using AETVgen.py script",     target)
            self._pp("# with the following parameters:",                      target)
            self._pp("#",                                                     target)
            for (key, value) in sorted(self.param.items(), key=lambda key_value: key_value[0]):
                if key not in ignoreParams:
                    self._pp("# {0:15} {1}".format(key, value), target)
            self._pp("#",                                                     target)
            self._pp("# ===================================================", target)

    def _pp(self, text, target=0):
        """ Pretty printing """
        if self.param['stdout'] == True:
            if target == self.FILE_ALL:
                tgt = "pdi, sdi, do"
            elif target == self.FILE_INPUT:
                tgt = "pdi, sdi"
            elif target == self.FILE_OUTPUT:
                tgt = "do"
            elif target == self.FILE_PUBLIC:
                tgt = "pdi, do"
            elif target == self.FILE_PUBLIC_INPUT:
                tgt = "pdi"
            elif target == self.FILE_SECRET:
                tgt = "sdi"
            elif target == self.FILE_C:
                tgt = "cfile"
            elif target == self.FILE_VHDL:
                tgt = "hdl"
            else:
                tgt = ""
            print ("{0:12} | {1}".format(tgt, text) )
        elif target == self.FILE_ALL:
            self.pdi.write(text + "\n")
            self.sdi.write(text + "\n")
            self.do.write(text + "\n")
        elif target == self.FILE_INPUT:
            self.pdi.write(text + "\n")
            self.sdi.write(text + "\n")
        elif target == self.FILE_OUTPUT:
            self.do.write(text + "\n")
        elif target == self.FILE_PUBLIC:
            self.pdi.write(text + "\n")
            self.do.write(text + "\n")
        elif target == self.FILE_PUBLIC_INPUT:
            self.pdi.write(text + "\n")
        elif target == self.FILE_SECRET:
            self.sdi.write(text + "\n")
        elif target == self.FILE_C:
            self.cfile.write(text + "\n")
        elif target == self.FILE_VHDL:
            self.hdl.write(text + "\n")
        else:
            raise ValueError("Invalid write target")
