-------------------------------------------------------------------------------
--! @file       AEAD_Core_Ent.vhd
--! @brief      Entity of authenticated encryption unit.
--!             User should modification to the default generics based on the
--!             implemented cipher.
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2015 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity AEAD_Core is
    generic (
        G_W                      : integer := 32;   --! Public data width (bits)
        G_SW                     : integer := 32;   --! Secret data width (bits)
        G_NPUB_SIZE              : integer := 128;  --! IV or Nonce size (bits)
        G_NSEC_ENABLE            : integer := 0;    --! Enable NSEC port
        G_NSEC_SIZE              : integer := 1;    --! NSEC width (bits)        
        G_ABLK_SIZE              : integer := 128;  --! Authenticated Data Block size (bits)
        G_DBLK_SIZE              : integer := 128;  --! Data Block size (bits)
        G_KEY_SIZE               : integer := 128;  --! Key size (bits)
        G_RDKEY_ENABLE           : integer := 0;    --! Enable rdkey port (also disables key port)
        G_RDKEY_SIZE             : integer := 1;    --! Roundkey size (bits)        
        G_TAG_SIZE               : integer := 128;  --! Tag size (bits)
        G_BS_BYTES               : integer := 4;    --! The number of bits required to hold block size expressed in bytes = log2_ceil(max(G_ABLK_SIZE,G_DBLK_SIZE)/8)
        G_LOADLEN_ENABLE         : integer := 0;    --! Enable load length section
        G_PAD                    : integer := 0;    --! Enable padding
        G_PAD_STYLE              : integer := 1;    --! Padding mode
        G_PAD_AD                 : integer := 1;    --! (G_PAD's sub option) Enable AD Padding
        G_PAD_D                  : integer := 1;    --! (G_PAD's sub option) Enable Data padding
        G_CTR_AD_SIZE            : integer := 16;   --! Maximum AD len value
        G_CTR_D_SIZE             : integer := 16;   --! Maximum data len value
        G_PLAINTEXT_MODE         : integer := 0;    --! Plaintext Mode
        G_CIPHERTEXT_MODE        : integer := 0;    --! Ciphertext mode
        G_REVERSE_DBLK           : integer := 0     --! Reverse block order (for message only)
    );
    port (
        --! Global signals
        clk                     :   in  std_logic;
        rst                     :   in  std_logic;

        --! Data in signals
        pdi                     :   in  std_logic_vector(G_W            -1 downto 0);
        pdi_valid               :   in  std_logic;
        pdi_ready               :   out std_logic;

        --! Key signals
        sdi                     :   in  std_logic_vector(G_SW           -1 downto 0);
        sdi_valid               :   in  std_logic;
        sdi_ready               :   out std_logic;

        --! Data out signals
        do                      :   out std_logic_vector(G_W            -1 downto 0);
        do_ready                :   in  std_logic;
        do_valid                :   out std_logic;

        --! FIFO signals
        bypass_fifo_wr          :   out std_logic;
        bypass_fifo_rd          :   out std_logic;
        bypass_fifo_full        :   in  std_logic;
        bypass_fifo_empty       :   in  std_logic;
        bypass_fifo_data        :   in  std_logic_vector(G_W            -1 downto 0);
        aux_fifo_din            :   out std_logic_vector(G_W            -1 downto 0);
        aux_fifo_ctrl           :   out std_logic_vector(4              -1 downto 0);
        aux_fifo_dout           :   in  std_logic_vector(G_W            -1 downto 0);
        aux_fifo_status         :   in  std_logic_vector(3              -1 downto 0)
    );
end AEAD_Core;
